/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * --------------------------------------------------------------------------*/
#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdmsql.h"

/*! \file perf06.c
 * \brief Source code for the RDM Perf06 RDMSQL example
 */
 
/*!
 * \mainpage Perf06_RDMSQL
 * \author Daigoro F. Toyama, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf06 RDMSQL
 *
 * The Perf06 example performs a number of atomic database operations
 * on a very simple database schema using the RDM SQL APIs. Different
 * transaction sizes (i.e. different number of rows inserted as part of
 * each transaction) to demonstrate the performance difference when using
 * small vs. large transactions.
 *
 * Each time you run this example
 *   * The database is initialized (removing all existing data)
 *   * 50,000 new records are created
 *   * Each record is deleted
 *   * 10,000 new records are created
 *   * Each record is deleted
 *   * 5,000 new records are created
 *   * Each record is deleted
 *   * 1,000 new records are created
 *   * Each record is deleted
 *   * 500 new records are created
 *   * Each record is deleted
 *   * 100 new records are created
 *   * Each record is deleted
 *   * Database is cleaned up
 *
 * \section Perf06_db Database
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file perf06.sdl. The example utilizes 
 * the SQL DDL syntax and specifies an on-disk database with a single 
 * table (simple).  The table has a single column (int_col).  There are no 
 * index definitions for this database.
 *
 * \section RDM SQL Functions Used
 * The following RDM functions are illustrated:
 *   * rsqlTFSInitEx/rsqlTFSTerm
 *   * rsqlAllocConn/rsqlFreeConn
 *   * rsqlOpenDB/rsqlCloseDB
 *   * rsqlInitDB
 *   * rsqlAllocStmt/rsqlFreeStmt
 *   * rsqlPrepare
 *   * rsqlBindParam
 *   * rsqlExecute
 *   * rsqlExecDirect
 *   * rsqlFetch
 *   * rsqlTransStart
 *   * rsqlTransCommit
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */

/*! \brief Create a test database
 *
 */
static RSQL_ERRCODE create_database(void)
{
    RSQL_ERRCODE rsqlError;
    HCONN        hDbc;
    HSTMT        hStmt;
    int16_t      ii;

    const char *cmds[] = {
        "CREATE DATABASE perf06_db",
        "CREATE TABLE simple (int_col INTEGER)",
        "COMMIT"
    };

    (void) rsqlAllocConn(&hDbc);
    (void) rsqlAllocStmt(hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    (void) rsqlExecDirect(hStmt, "DROP DATABASE perf06_db");
    (void) rsqlExecDirect(hStmt, "COMMIT");

    for (ii = 0; ii < sizeof(cmds) / sizeof(cmds[0]); ii++)
    {
        rsqlError = rsqlExecDirect(hStmt, cmds[ii]);
        if (rsqlError != errSUCCESS)
            break;
    }

    if (rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error creating the perf06_db database\n");
        return rsqlError;
    }

    /* Close the newly created database */
    (void) rsqlCloseDB(hDbc, "perf06_db");

    (void) rsqlFreeStmt(hStmt);
    (void) rsqlFreeConn(hDbc);

    return errSUCCESS;
}

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void display_elapsed_time(
    const char *msg,         /**< [in] Message to be display along with the elapsed time */
    uint64_t    start_time,  /**< [in] Start of elapsed time to display */
    uint64_t    end_time)    /**< [in] End of elapsed */
{
    uint64_t elapsed_time = end_time - start_time;
    printf("%s" UINT64_SPEC_SIZE("5") " millisecs\n", msg, elapsed_time/1000000);
}

/*! \brief Initialize the RDM SQL library for use in the perf06 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use 
 * the single-user transactional DIRECT-LINK implementation.  It also creates
 * an RDM runtime task and opens the "perf06_db" database in exclusive ('x') mode.
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static RSQL_ERRCODE initialize(
    HCONN    *hDBC)   /**< [out] Pointer to the connection handle to be initialized */
{
    RSQL_ERRCODE rsqlError;
    uint64_t     start_time;
    uint64_t     end_time;

    *hDBC = NULL;

    /* Initialize the TFS to use the DIREC-LINK configuration (TFST) */    
    start_time = psp_timeNanoSecs();

    rsqlError = create_database();
    if(rsqlError != errSUCCESS)
        return rsqlError;

    rsqlError = rsqlTFSInitEx(NULL, NULL, TFS_TYPE_TFS);
    if(rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error initializing the TFS\n");
        return rsqlError;
    }

    /* Allocate the connection handle */
    rsqlError = rsqlAllocConn(hDBC);
    if(rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error allocating a connection handle\n");
        return rsqlError;
    }

    /* Open the perf06_db database in exclusive mode (does not require locks or tranactions) */
    rsqlError = rsqlOpenDB(*hDBC, "perf06_db", "x");
    if(rsqlError != errSUCCESS)
    {
        fprintf(stderr, "Error opening the perf06_db database\n");
        return rsqlError;
    }

    /* Initialize the database (creates new database files truncating any existing data) */    
    rsqlError = rsqlInitDB(*hDBC, "perf06_db");
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occurred initializing database");
        return rsqlError;
    }

    end_time = psp_timeNanoSecs();  
    
    /* Display information */
    printf("Initializing with DIREC-LINK configuration\n");
    display_elapsed_time("Preparing a new database:\t\t\t\t", start_time, end_time);
    printf("\tOne, simple record type.\n");
    printf("\tOn disk\n");
    printf("\tOpened exclusively\n");

    return rsqlError;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM connection
 * handle used in the perf06 example.  It also terminates the DIRECT-LINK TFS
 * used in the example. 
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static void cleanup(
    HCONN hDBC)   /**< [in] Connection handle to be closed/cleaned up */

{
    uint64_t     start_time;
    uint64_t     end_time;

    start_time = psp_timeNanoSecs();

    if (hDBC)
    {
        /* Close any open databases */
        rsqlCloseDB(hDBC, "perf06_db");

        /* Free the connection handle */
        rsqlFreeConn(hDBC);
    }

    /* Terminate the DIRECT-LINK TFS */
    rsqlTFSTerm();

    end_time = psp_timeNanoSecs();

    display_elapsed_time("Cleanup:\t\t\t\t\t\t", start_time, end_time);
}

/*! \brief Insert 50,000 simple records into the perf06_db database
 *
 * This function adds 50,000 simple records into the on-disk perf database.  The
 * simple table contains only a single integer column with no index.  The function will
 * calculate and display the total time elapsed while inserting the records
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
static int32_t add_records(
    HCONN   hDBC,       /**< [in] RDM SQL Connection handle */
    int32_t transSize)  /**< [in] Number of rows to add in one transaction */
{
    int32_t       rdm_status = S_OKAY;
    int32_t       ii;
    int32_t       int_col = 0;
    int32_t       count = 0;
    uint64_t      start_time;
    uint64_t      end_time;
    char          msg[100];
    HSTMT         hSTMT;
    RSQL_ERRCODE  rsqlError;

    start_time = psp_timeNanoSecs();

    /* allocates insert statement */
    rsqlError = rsqlAllocStmt(hDBC, &hSTMT);
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occured allocating insert statement");
        return rsqlError;
    }

    /* Prepare insert statement */
    rsqlError = rsqlPrepare(hSTMT,  "insert into simple values(?)");
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred preparing the insert statement");
        return rsqlError;
    }

    rsqlError = rsqlBindParam(hSTMT, 1, tINTEGER, &int_col, 0);
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred binding the insert statement parameter");
        return rsqlError;
    }

    /* Loop to add 50,000 new simple record */
    for(ii = 0; ii < 50000; ii++)
    {
        /* Set values for the new record */
        int_col = 1;

        if (count == 0)
            (void) rsqlTransStart(hDBC, "Perf06");

        /* Add the record to the database */
        rdm_status = rsqlExecute(hSTMT);
        if (rsqlError != errSUCCESS)
        {
            rsqlFreeStmt(hSTMT);
            printf("Error has occurred exeucting the insert statement");
            return rsqlError;
        }

        if (++count == transSize)
        {
            (void) rsqlTransCommit(hDBC);
            count = 0;
        }
    }

    if (count > 0)
        (void) rsqlTransCommit(hDBC);

    rsqlFreeStmt(hSTMT);
    end_time = psp_timeNanoSecs();

    sprintf(msg, "Add 50,000 records with %5d records/transaction:\t",
            transSize);
    display_elapsed_time(msg, start_time, end_time);
    return rdm_status;
}

/*! \brief Delete all simple records from the database
 *
 * This function will delete all simple records stored in the perf06_db database.
 * The function will calculate and display the total time elapsed while 
 * deleting the records.
 *
 * @return Returns a RSQL_ERRCODE value (errSUCCESS if successful)
 */
int32_t delete_records(
    HCONN hDBC)         /**< [in]
 RDM task initialized with the perf06_db database open and as the current database */
{
    uint64_t      start_time;
    uint64_t      end_time;
    HSTMT         hSTMT;
    RSQL_ERRCODE  rsqlError;

    start_time = psp_timeNanoSecs();
    
    /* allocates delete statement */
    rsqlError = rsqlAllocStmt(hDBC, &hSTMT);
    if (rsqlError != errSUCCESS)
    {
        printf("Error has occured allocating delete statement");
        return rsqlError;
    }

    /* Execute delete statement */
    rsqlError = rsqlExecDirect(hSTMT,  "delete from simple");
    if (rsqlError != errSUCCESS)
    {
        rsqlFreeStmt(hSTMT);
        printf("Error has occurred executing the delete statement\n");
        return rsqlError;
    }

    rsqlFreeStmt(hSTMT);
    end_time = psp_timeNanoSecs();

    display_elapsed_time("Deleting 50,000 records:\t\t\t\t", start_time, end_time);
    return rsqlError;
}

/*! \brief Main function for perf06 example
 *
 * The function initializes the RDM SQL environment and runs the create, read,
 * update, and delete operations.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    int32_t        rdm_status;
    HCONN          hDBC;
    static int32_t transactionSizes[] = { 50000, 10000, 5000, 1000, 500, 100 };
    int32_t        ii;

    /* Display the example title */
    printf("%s", RDM_EXAMPLE_DESC("Perf06"));

    /* Initialize the TFS, task, and open/initialize the database */
    rdm_status = initialize(&hDBC);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Application is aborting due to initialization error\n");
        cleanup(hDBC);
        return EXIT_FAILURE;
    }

    for (ii = 0; ii < RLEN(transactionSizes); ii++)
    {
        /* Add 50,000 records to the database */
        rdm_status = add_records(hDBC, transactionSizes[ii]);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Application is aborting due to error adding records\n");
            cleanup(hDBC);
            return EXIT_FAILURE;
        }

        /* Remove all of the records in the database */
        rdm_status = delete_records(hDBC);
        if(rdm_status != S_OKAY)
        {
            fprintf(stderr, "Application is aborting due to error deleting records\n");
            cleanup(hDBC);
            return EXIT_FAILURE;
        }
    }

    cleanup(hDBC);

    printf("\nPress Enter to terminate");
    getchar();

    return EXIT_SUCCESS;
}
